import axios from 'axios'
import get from 'lodash/get'
import sortBy from 'lodash/sortBy'
import uniqBy from 'lodash/uniqBy'

export const CONFIG = {
  servers: [
    {
      name: "dev.funkwhale.audio",
      protocol: "https",
      apiPath: "/api/v4/",
      type: 'gitlab',
      params: {
        per_page: 15,
        'not[labels]': 'Status: Draft',
      },
      types: {
        labels: {
          'Status: input wanted': 'Engage in the discussion',
          'Status: Need triage': 'Categorize the task',
          'Status: Ready': 'Complete the task',
          'Status: Needs reproduction': 'Reproduce the bug',
          'Status: Review wanted': 'Give your feedback',
        }
      },
    }
  ],
  serversByName: {},
  repositories: [
    {
      server: "dev.funkwhale.audio",
      namespace: "funkwhale/tasks",
      label: 'Funkwhale / Tasks',
      projectId: 185,
    },
    {
      server: "dev.funkwhale.audio",
      namespace: "funkwhale/funkwhale",
      projectId: 17,
      label: 'Funkwhale',
      params: {
        milestone: 'Any',
      },
      categories: [
        {
          label: 'Documentation',
          filters: {
            labels: 'Area: Documentation',
            'not[labels]': 'Type: New feature',
          }
        },
        {
          label: 'Web UI',
          filters: {
            labels: 'Area: Frontend',
            'not[labels]': 'Area: Backend',
          }
        },
        {
          label: 'API',
          filters: {
            labels: 'Area: Backend',
            'not[labels]': 'Area: Frontend',
          }
        },
        {
          label: 'Support',
          filters: {
            labels: 'Area: Support'
          }
        },
      ]
    },
  ],
  repositoriesById: {}
}

export const SERVER_TYPES = {
  gitlab: {
    getIssuesListUrl ({server, repository}) {
      return `${server.baseApiUrl}projects/${repository.projectId}/issues`
    },
    getIssuesListParams ({server, repository, categoryFilters, additional}) {
      let serverTypeDefaults = {per_page: 30, state: 'opened', with_labels_details: true}
      let serverDefaults = get(server, 'params', {})
      let repositoryDefaults = get(repository, 'params', {})
      additional = additional || {}
      let concatenatedFilters = this.concatFilters([serverTypeDefaults,serverDefaults, repositoryDefaults, categoryFilters, additional])
      return {...serverTypeDefaults,...serverDefaults, ...repositoryDefaults, ...categoryFilters, ...additional, ...concatenatedFilters}
    },
    concatFilters (filtersGroups) {
      let concatFields = ['labels', 'not[labels]']
      let values = {}
      concatFields.forEach(f => {
        filtersGroups.forEach(g => {
          if (g[f]) {
            values[f] = values[f] || []
            values[f].push(g[f])
          }
        })
        if (values[f]) {
          values[f] = values[f].join(',')
        }
      })
      return values


    },
    async getIssuesList ({server, repository, categoryFilters, params}) {
      let query = this.getIssuesListParams({server, repository, categoryFilters, additional: params})
      let url = this.getIssuesListUrl({server, repository})
      let response = await axios.get(url, {params: query})
      return this.parseIssuesListResponse({repository, response})
    },
    parseIssuesListResponse({repository, response}) {
      return {
        repository,
        objs: response.data,
        pagination: {
          current: response.headers['x-page'],
          next: response.headers['x-next-page'],
          prev: response.headers['x-prev-page'],
          totalPages: parseInt(response.headers['x-total-pages']),
          total: parseInt(response.headers['x-total'],)
        },
        status: response.status,
        response
      }
    },
    getSortValue (obj, sortKey) {
      if (sortKey === 'popularity') {
        return obj.upvotes
      }
      let value = obj[sortKey]
      if (sortKey === 'due_date') {
        value = value || ''
      }
      return value
    }
  }
}

CONFIG.servers.forEach((s) => {
  s.baseApiUrl = `${s.protocol}://${s.name}${s.apiPath}`
  s.handlers = SERVER_TYPES[s.type]
  CONFIG.serversByName[s.name] = s
})

CONFIG.repositories.forEach((r) => {
  r.serverObj = CONFIG.serversByName[r.server]
  r.id = `${r.serverObj.name}/${r.namespace}/${r.id}`
})

export async function mergeIssuesList(promises, options) {
  const data = {
    parsedResponses: [],
    objs: [],
    total: 0,
  }
  let sortedObjs = []
  for (var i = 0; i < promises.length; i++) {
    let promiseData = promises[i]
    let responseData = await promiseData.promise
    data.parsedResponses.push(responseData)
    data.total += responseData.pagination.total
    responseData.objs.forEach((o) => {
      o.repositoryObj = promiseData.repository
      o.categoryObj = promiseData.category
      o.uid = `${o.repositoryObj.baseApiUrl}/${o.iid}`
      let sortValue = o.repositoryObj.serverObj.handlers.getSortValue(o, options.orderBy)
      if (sortValue !== undefined) {
        sortedObjs.push({sortValue, obj: o})
      }
    })
  }
  sortedObjs = uniqBy(sortedObjs, 'obj.uid')
  // now, sorting everything, yay, fun
  sortedObjs = sortBy([...sortedObjs], ['sortValue'])
  if (options.sort === 'desc') {
    sortedObjs.reverse()
  }
  data.objs = sortedObjs.map((sortedObj) => {
    return sortedObj.obj
  })
  return data
}
export async function getIssues(params) {
  let promises = []
  CONFIG.repositories.forEach((r) => {
    let categories = r.categories
    if (!categories) {
      categories = [r]
    }
    categories.forEach((c) => {
      let filters = c.filters || {}
      let promise = r.serverObj.handlers.getIssuesList({server: r.serverObj, repository: r, categoryFilters: filters, params})
      promises.push({repository: r, promise, category: c})
    })
  })
  let options = {
    sort:  params.sort || 'desc',
    orderBy:  params.order_by,
  }
  return await mergeIssuesList(promises, options)
}
